from django_filters.filters import Filter
from typing import List, Optional
from django.forms import Form


class FormFieldNotFoundException(Exception):

    def __init__(self, field_name: str):
        super().__init__(f'Campo "{field_name}" no encontrado en el formulario')


class Column:

    def __init__(self, field_name: str, css_class: str = ""):
        self.field_name = field_name
        self.css_class = css_class

    def render_label(self, field: Filter) -> str:
        widget_attr_id = field.widget.attrs.get('id', "")
        return f'<label for="{widget_attr_id}" class="form-label">{field.label}</label>'

    def render(self, field: Filter, field_value: Optional[str]) -> str:
        widget_render = field.widget.render(name=self.field_name, value=field_value)
        return f'<div class="{self.css_class}"> { self.render_label(field) } {widget_render} </div>'


class Row:

    def __init__(self, fields: List[Column], css_class: str = ""):
        self.fields = fields
        self.css_class = css_class

    def render(self, form: Form) -> str:
        row_html = f'<div class="row {self.css_class}">'
        for column in self.fields:
            field_name = column.field_name
            field = form.fields.get(field_name)
            field_value = form.data.get(field_name, '')
            if not field:
                raise FormFieldNotFoundException(field_name)
            row_html += column.render(field, field_value)
        row_html += '</div>'
        return row_html


class FilterLayout:

    def __init__(self, rows: List[Row] = None):
        self.rows = rows if rows is not None else []

    def render(self, form: Form) -> str:
        return '\n'.join(row.render(form) for row in self.rows)



from typing import Any, Dict
from django.views import View
from django.core.paginator import Paginator
from abc import ABC, abstractmethod
from django.shortcuts import render
from .mixins import (
    QueryFilterOrderMixin,
    InvalidQueryFilterException,
    InvalidOrderDirectionException,
    InvalidOrderFieldException
)

MESSAGE_TYPE_WARNING = 'warning'
MESSAGE_TYPE_DANGER = 'danger'
MESSAGE_BI_ICONS = {
    'warning': 'bi bi-exclamation-octagon-fill text-warning',
    'danger': 'bi bi-exclamation-triangle text-danger'
}
MESSAGE_DEFAULT_ICON = 'bi bi-info-circle text-primary'


class InvalidPageNumberException(Exception):
    def __init__(self, current_page: int, max_page: int):
        super().__init__(f'El número de página {current_page} es inválido, el rango permitido es de 1 a {max_page}')


class PaginatorView(ABC, View, QueryFilterOrderMixin):

    default_value: int = 5
    default_page: int = 1
    default_per_page: int = 5
    default_on_each_side: int = 2
    default_on_ends: int = 0
    max_per_page: int = 100
    export_url: str = None
    import_url: str = None

    @property
    @abstractmethod
    def template_name(self) -> str:
        pass

    def get_integer_query_param(self, name: str, default_value: int, max_value: int = None) -> int:
        try:
            value = int(self.request.GET.get(name, default_value))
        except (ValueError, TypeError):
            value = default_value
        if max_value and value > max_value:
            value = max_value
        return value

    def get_page_obj(self, query):
        per_page = self.get_integer_query_param('per_page', self.default_per_page, self.max_per_page)
        page_number = self.get_integer_query_param('page', self.default_page)
        paginator = Paginator(query, per_page)
        if page_number < 1:
            page_number = 1
        if page_number > paginator.num_pages:
            raise InvalidPageNumberException(page_number, paginator.num_pages)
        page_obj = paginator.get_page(page_number)
        pages = paginator.get_elided_page_range(
            number=page_number,
            on_each_side=self.default_on_each_side,
            on_ends=self.default_on_ends
        )
        return page_obj, pages

    def get_context_error(self, exception: Exception, message_type: str) -> Dict[str, str]:
        return {
            'type': message_type,
            'icon': MESSAGE_BI_ICONS.get(message_type, MESSAGE_DEFAULT_ICON),
            'message': str(exception)
        }

    def build_context(self, request) -> Dict[str, Any]:
        pages = []
        page_obj = None
        page_error = None
        try:
            queryset = self.get_queryset()
            queryset = self.apply_filterset_to_queryset(queryset, request.GET)
            queryset = self.apply_order_to_queryset(queryset, request.GET)
            page_obj, pages = self.get_page_obj(queryset)
        except (InvalidQueryFilterException, InvalidOrderDirectionException, InvalidOrderFieldException) as e:
            page_error = self.get_context_error(e, MESSAGE_TYPE_WARNING)
        except (InvalidPageNumberException) as e:
            page_error = self.get_context_error(e, MESSAGE_TYPE_DANGER)
        except Exception as e:
            raise e
        return {
            'page_error': page_error,
            'page_obj': page_obj,
            'pages': pages,
            'filterset': self.filterset,
            'filters': self.applied_filters,
            'filters_errors': self.filters_errors,
            'order': self.applied_order,
            'defaults': self.get_defaults(),
            'export_url': self.export_url,
            'import_url': self.import_url
        }

    def get(self, request, *args, **kwargs):
        context = self.build_context(request)
        return render(request, self.template_name, context)



from abc import ABC, abstractmethod
from django.db.models import Model
from .query_filter_mixin import QueryFilterMixin
from .query_order_mixin import QueryOrderMixin


class QueryFilterOrderMixin(ABC, QueryFilterMixin, QueryOrderMixin):

    @property
    @abstractmethod
    def model(self) -> Model:
        raise NotImplementedError

    def __init__(self):
        self.init_filterset_defaults()
        self.init_order_defaults()

    def get_queryset(self):
        return self.model.objects.all()

    def get_complete_queryset(self, params: dict):
        queryset = self.get_queryset()
        queryset = self.apply_filterset_to_queryset(queryset, params)
        queryset = self.apply_order_to_queryset(queryset, params)
        return queryset

    def get_defaults(self):
        return {
            'order_by': self.default_order_by,
            'order_direction': self.default_order_direction,
            'fields_order': self.allowed_fields_order,
            'field_time': self.default_field_time
        }



from typing import Dict
from django.db.models import QuerySet

DEFAULT_ID_FIELD = 'id'
DEFAULT_TIME_FIELD = 'created_at'
DIRECTION_ASC = 'ASC'
DIRECTION_DESC = 'DESC'
DIRECTIONS = {DIRECTION_ASC, DIRECTION_DESC}


class InvalidOrderDirectionException(Exception):
    def __init__(self, order_direction: str):
        super().__init__(f'La dirección de ordenamiento "{order_direction}" es inválida, debe ser "ASC" o "DESC".')


class InvalidOrderFieldException(Exception):
    def __init__(self, field_name: str, allowed_fields: set):
        allowed_fields_str = ', '.join(sorted(allowed_fields))
        super().__init__(f'El campo de ordenamiento "{field_name}" es inválido. Campos permitidos: {allowed_fields_str}.')


class QueryOrderMixin:
    default_order_by: str = DEFAULT_ID_FIELD
    default_order_direction: str = DIRECTION_DESC
    default_field_time: str = DEFAULT_TIME_FIELD
    allowed_fields_order: set = {DEFAULT_ID_FIELD, DEFAULT_TIME_FIELD}

    def init_order_defaults(self):
        self.applied_order = self.validate_order_params(self.default_order_by, self.default_order_direction)

    def validate_order_params(self, order_by_field: str, order_direction: str) -> Dict[str, str]:
        """
        Valida los parámetros de ordenamiento y construye la configuración de orden.

        Args:
            order_by_field (str): Campo por el cual se ordenará.
            order_direction (str): Dirección del ordenamiento ("ASC" o "DESC").

        Returns:
            Dict[str, str]: Un diccionario con los detalles del ordenamiento.

        Raises:
            InvalidOrderFieldException: Si el campo de ordenamiento no es válido.
            InvalidOrderDirectionException: Si la dirección de ordenamiento no es válida.
        """
        if order_direction not in DIRECTIONS:
            raise InvalidOrderDirectionException(order_direction)
        if order_by_field not in self.allowed_fields_order:
            raise InvalidOrderFieldException(order_by_field, self.allowed_fields_order)

        order_by = f"-{order_by_field}" if order_direction == DIRECTION_DESC else order_by_field
        return {
            'order_by': order_by,
            'by': order_by_field,
            'direction': order_direction
        }

    def apply_order_to_queryset(self, queryset: QuerySet, order_params: dict) -> QuerySet:
        order_by_field = order_params.get('order_by', self.default_order_by)
        order_direction = order_params.get('order_direction', self.default_order_direction)
        self.applied_order = self.validate_order_params(order_by_field, order_direction)
        return queryset.order_by(self.applied_order['order_by'])



from .query_order_mixin import (
    QueryOrderMixin,
    DIRECTION_ASC,
    DIRECTION_DESC,
    DIRECTIONS,
    InvalidOrderDirectionException,
    InvalidOrderFieldException
)
from .query_filter_mixin import QueryFilterMixin, InvalidQueryFilterException
from .query_filter_order_mixin import QueryFilterOrderMixin

__all__ = [
    'QueryOrderMixin',
    'DIRECTION_ASC',
    'DIRECTION_DESC',
    'DIRECTIONS',
    'InvalidOrderDirectionException',
    'InvalidOrderFieldException',
    'QueryFilterMixin',
    'InvalidQueryFilterException',
    'QueryFilterOrderMixin'
]



from .paginator_view import PaginatorView
from .filter_layout import (
    FormFieldNotFoundException,
    Column,
    Row,
    FilterLayout
)

__all__ = [
    'PaginatorView',
    'Column',
    'Row',
    'FilterLayout',
    'FormFieldNotFoundException'
]



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='django-pagination-utils',
    version='0.1.1',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Django>=5.0.0',
        'django-filter>=2.4.0'
    ],
    setup_requires=['wheel'],
    python_requires='>=3.10',
    license='MIT',
    description='Django paginations utils',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cristhianCEM/django-paginations-utils',
    author='cristianCEM',
    author_email='cristianbcer27@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 5.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3'
    ],
    keywords=['Django', 'Paginations', 'Utils', 'Filters']
)



from abc import ABC, abstractmethod
from django_filters import FilterSet
from django.db.models import QuerySet


class InvalidQueryFilterException(Exception):
    def __init__(self):
        super().__init__(f'Los filtros ingresados son inválidos o tienen errores.')


class QueryFilterMixin(ABC):

    @property
    @abstractmethod
    def filterset_class(self) -> FilterSet:
        raise NotImplementedError(
            f"El atributo 'filterset_class' no ha sido implementado en la clase {self.__class__.__name__}. "
            "Define una subclase de 'FilterSet' para usar este mixin."
        )

    def init_filterset_defaults(self):
        self.filterset = None
        self.applied_filters: dict = {}
        self.filters_errors: dict = {}

    def add_applied_filters(self):
        """
        Registra los filtros aplicados exitosamente en `applied_filters`.
        """
        cleaned_data = self.filterset.form.cleaned_data
        for key, value in cleaned_data.items():
            if not value:
                continue
            field = self.filterset.form.fields.get(key)
            if not field:
                continue
            self.applied_filters[key] = {
                'label': field.label,
                'value': value
            }

    def add_filters_errors(self):
        """
        Registra los errores de los filtros en `filters_errors`.
        """
        for key in self.filterset.filters:
            field = self.filterset.form.fields.get(key)
            errors = self.filterset.errors.get(key)
            if not field or not errors:
                continue
            self.filters_errors[key] = {
                'label': field.label,
                'errors': errors
            }

    def apply_filterset_to_queryset(self, queryset: QuerySet, filter_params: dict) -> QuerySet:
        """
        Aplica los filtros definidos en `filterset_class` al queryset.

        Args:
            filter_params: Parámetros de los filtros a aplicar.
            queryset: Queryset a filtrar.

        Returns:
            Queryset filtrado según los parámetros especificados.

        Raises:
            InvalidQueryFilterException: Si los filtros especificados no son válidos.
        """
        self.init_filterset_defaults()
        self.filterset = self.filterset_class(filter_params, queryset=queryset)
        queryset = self.filterset.qs
        self.add_applied_filters()
        if not self.filterset.is_valid():
            self.add_filters_errors()
            raise InvalidQueryFilterException()
        return queryset
